@@ -13,9 +13,7 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library; if not, see <http://www.gnu.org/licenses/>.
  *
  * Author: Ryan Lortie <desrt@desrt.ca>
  */
@@ -61,6 +59,7 @@ typedef struct
 static void dconf_writer_iface_init (DConfDBusWriterIface *iface);
 
 G_DEFINE_TYPE_WITH_CODE (DConfWriter, dconf_writer, DCONF_DBUS_TYPE_WRITER_SKELETON,
+                         G_ADD_PRIVATE (DConfWriter)
                          G_IMPLEMENT_INTERFACE (DCONF_DBUS_TYPE_WRITER, dconf_writer_iface_init))
 
 static void
@@ -345,7 +344,7 @@ dconf_writer_iface_init (DConfDBusWriterIface *iface)
 static void
 dconf_writer_init (DConfWriter *writer)
 {
-  writer->priv = G_TYPE_INSTANCE_GET_PRIVATE (writer, DCONF_TYPE_WRITER, DConfWriterPrivate);
+  writer->priv = dconf_writer_get_instance_private (writer);
   writer->priv->basepath = g_build_filename (g_get_user_config_dir (), "dconf", NULL);
   writer->priv->native = TRUE;
 }
@@ -381,8 +380,6 @@ dconf_writer_class_init (DConfWriterClass *class)
                                    g_param_spec_string ("name", "name", "name", NULL,
                                                         G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT_ONLY |
                                                         G_PARAM_WRITABLE));
-
-  g_type_class_add_private (class, sizeof (DConfWriterPrivate));
 }
 
 void
