@@ -149,6 +149,12 @@ dconf_engine_handle_dbus_signal (GBusType     bus_type,
 static void
 test_creation_error (void)
 {
+  if (g_getenv ("DISPLAY") == NULL || g_strcmp0 (g_getenv ("DISPLAY"), "") == 0)
+    {
+      g_test_skip ("FIXME: D-Bus tests do not work on CI at the moment");
+      return;
+    }
+
   /* Sync with 'error' */
   if (g_test_trap_fork (0, 0))
     {
@@ -230,6 +236,12 @@ test_sync_call_success (void)
   gchar *system_id;
   GVariant *reply;
 
+  if (g_getenv ("DISPLAY") == NULL || g_strcmp0 (g_getenv ("DISPLAY"), "") == 0)
+    {
+      g_test_skip ("FIXME: D-Bus tests do not work on CI at the moment");
+      return;
+    }
+
   reply = dconf_engine_dbus_call_sync_func (G_BUS_TYPE_SESSION,
                                             "org.freedesktop.DBus", "/", "org.freedesktop.DBus", "ListNames",
                                             g_variant_new ("()"), G_VARIANT_TYPE ("(as)"), &error);
@@ -271,6 +283,12 @@ test_sync_call_error (void)
   GError *error = NULL;
   GVariant *reply;
 
+  if (g_getenv ("DISPLAY") == NULL || g_strcmp0 (g_getenv ("DISPLAY"), "") == 0)
+    {
+      g_test_skip ("FIXME: D-Bus tests do not work on CI at the moment");
+      return;
+    }
+
   /* Test receiving errors from the other side */
   reply = dconf_engine_dbus_call_sync_func (G_BUS_TYPE_SESSION,
                                             "org.freedesktop.DBus", "/", "org.freedesktop.DBus", "GetId",
@@ -347,6 +365,12 @@ test_async_call_success (void)
 {
   gint i;
 
+  if (g_getenv ("DISPLAY") == NULL || g_strcmp0 (g_getenv ("DISPLAY"), "") == 0)
+    {
+      g_test_skip ("FIXME: D-Bus tests do not work on CI at the moment");
+      return;
+    }
+
   for (i = 0; i < 1000; i++)
     {
       DConfEngineCallHandle *handle;
@@ -375,6 +399,12 @@ test_async_call_error (void)
   GError *error = NULL;
   gboolean success;
 
+  if (g_getenv ("DISPLAY") == NULL || g_strcmp0 (g_getenv ("DISPLAY"), "") == 0)
+    {
+      g_test_skip ("FIXME: D-Bus tests do not work on CI at the moment");
+      return;
+    }
+
   handle = (gpointer) g_variant_type_new ("(s)");
 
   g_mutex_lock (&async_call_queue_lock);
@@ -398,6 +428,12 @@ test_sync_during_async (void)
   gboolean success;
   GVariant *reply;
 
+  if (g_getenv ("DISPLAY") == NULL || g_strcmp0 (g_getenv ("DISPLAY"), "") == 0)
+    {
+      g_test_skip ("FIXME: D-Bus tests do not work on CI at the moment");
+      return;
+    }
+
   handle = (gpointer) g_variant_type_new ("(s)");
   g_mutex_lock (&async_call_queue_lock);
   g_queue_push_tail (&async_call_success_queue, handle);
@@ -433,6 +469,12 @@ test_signal_receipt (void)
   gint status;
   guint id;
 
+  if (g_getenv ("DISPLAY") == NULL || g_strcmp0 (g_getenv ("DISPLAY"), "") == 0)
+    {
+      g_test_skip ("FIXME: D-Bus tests do not work on CI at the moment");
+      return;
+    }
+
   reply = dconf_engine_dbus_call_sync_func (G_BUS_TYPE_SESSION,
                                             "org.freedesktop.DBus", "/", "org.freedesktop.DBus", "AddMatch",
                                             g_variant_new ("(s)", "type='signal',interface='ca.desrt.dconf.Writer'"),
@@ -466,8 +508,7 @@ main (int argc, char **argv)
   if (!g_str_equal (DBUS_BACKEND, "/libdbus-1"))
     g_test_add_func (DBUS_BACKEND "/creation/error", test_creation_error);
 
-  // FIXME: This test is failing.
-  //g_test_add_func (DBUS_BACKEND "/sync-call/success", test_sync_call_success);
+  g_test_add_func (DBUS_BACKEND "/sync-call/success", test_sync_call_success);
   g_test_add_func (DBUS_BACKEND "/sync-call/error", test_sync_call_error);
   g_test_add_func (DBUS_BACKEND "/async-call/success", test_async_call_success);
   g_test_add_func (DBUS_BACKEND "/async-call/error", test_async_call_error);
